using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Text.RegularExpressions;
using MinorShift.Emuera.Sub;
using MinorShift.Emuera.GameData;
using MinorShift.Emuera.GameView;
using MinorShift.Emuera.GameData.Expression;
using MinorShift.Emuera.GameData.Variable;
namespace MinorShift.Emuera.GameProc
{

	internal sealed partial class Process
	{
		private sealed class ErbLoader
		{
			public ErbLoader(EmueraConsole main, VariableEvaluator var)
			{
				output = main;
				varData = var;
			}
			readonly EmueraConsole output;
			readonly VariableEvaluator varData;
			Dictionary<string, int> ignoredWarningCount = new Dictionary<string, int>();
			List<string> ignoredFNFWarningFileList = new List<string>();
			int ignoredFNFWarningCount = 0;

			int enabledLineCount = 0;
			int lineCount = 0;
			LabelDictionary labelDic;


			/// <summary>
			/// ̃t@Cǂ
			/// </summary>
			/// <param name="filepath"></param>
			public LabelDictionary LoadErbFiles(string erbDir, bool displayReport)
			{

				labelDic = new LabelDictionary();
				List<KeyValuePair<string, string>> erbFiles = Config.Instance.GetFiles(erbDir, "*.ERB");
				int first = Environment.TickCount;
				try
				{
					for (int i = 0; i < erbFiles.Count; i++)
					{
						string file = erbFiles[i].Value;
						string filename = erbFiles[i].Key;
#if DEBUG
						if (Config.Instance.DisplayReport)
							output.PrintLine(filename + "ǂݍݒEEE" + (Environment.TickCount - first).ToString() + ":");
#else
						if (displayReport)
							output.PrintLine(filename + "ǂݍݒEEE");
#endif
						System.Windows.Forms.Application.DoEvents();
						loadErb(file, filename);
					}
					checkScript();
				}
				catch (CodeEE e)
				{
					if (e.Position != null)
					{
						output.PrintLine(e.Position.Filename + "" + e.Position.LineNo.ToString() + "sڂŃG[܂");
						output.PrintLine(e.Position.RowLine);
					}
					output.PrintLine(e.Message);
					return null;
				}
				catch (Exception e)
				{
					output.PrintLine("\ȂG[܂");
					output.PrintLine(e.GetType().ToString() + ":" + e.Message);
					return null;
				}
				return labelDic;
			}

			/// <summary>
			/// t@Cǂ
			/// </summary>
			/// <param name="filepath"></param>
			private void loadErb(string filepath, string filename)
			{
				EraStreamReader eReader = new EraStreamReader();
				if (!eReader.Open(filepath, filename))
				{
					output.PrintLine(eReader.Filename + "̃I[vɎs܂");
					return;
				}
				try
				{
					string line = null;
					LogicalLine lastLine = new NullLine();
					FunctionLabelLine lastLabelLine = null;
					StringStream st = null;
					bool skip = false;
					string rowLine = null;
					while ((line = eReader.ReadLine()) != null)
					{
						lineCount++;
						if (line.Length == 0)
							continue;
						rowLine = line;
						st = new StringStream(line);
						TokenReader.SkipWhiteSpace(st);
						if (st.Current == ';')
						{
							if (!st.CurrentEqualTo(";!;"))
								continue;
							st.ShiftNext();
							st.ShiftNext();
							st.ShiftNext();
							TokenReader.SkipWhiteSpace(st);
							if (st.Current == ';')
								continue;
						}
						if (st.EOS)
							continue;
						line = st.Substring();
						if (Config.Instance.UseRenameFile)
							line = applyRename(line);
						st = new StringStream(line);
						ScriptPosition position = new ScriptPosition(eReader.Filename, eReader.LineNo, rowLine);
						if (st.Current == '[')
						{
							st.ShiftNext();
							if (TokenReader.GetNextTokenType(st) != TokenType.Identifer)
								output.PrintWarning("[]̎gsł", position, 1);
							string token = TokenReader.ReadSingleIdentifer(st);
							if ((token == null) || (st.Current != ']'))
								output.PrintWarning("[]̎gsł", position, 1);
							switch (token)
							{
								case "SKIPSTART":
									if (skip)
										output.PrintWarning("[SKIPSTART]dĎgpĂ܂", position, 1);
									skip = true;
									break;
								case "SKIPEND":
									if (!skip)
										output.PrintWarning("[SKIPSTART]ƑΉȂ[SKIPEND]ł", position, 1);
									skip = false;
									break;
								default:
									output.PrintWarning("FłȂvvZbTł", position, 1);
									break;
							}
							continue;
						}
						if (skip == true)
							continue;
						//܂ŃvvZbT

						if (st.Current == '#')
						{
							if (!(lastLine is FunctionLabelLine))
							{
								output.PrintWarning("֐錾̒ȊO#sgĂ܂", position, 1);
								continue;
							}
							FunctionLabelLine label = (FunctionLabelLine)lastLine;
							st.ShiftNext();
							string token = TokenReader.ReadSingleIdentifer(st);
							if (token == null)
							{
								output.PrintWarning("߂łȂ#sł", position, 1);
								continue;
							}
							if (Config.Instance.IgnoreCase)
								token = token.ToUpper();
							switch (token)
							{
								case "SINGLE": label.IsSingle = true; break;
								case "LATER": label.Priority++; break;
								case "PRI": label.Priority--; break;
								default:
									output.PrintWarning("߂łȂ#sł", position, 1);
									break;
							}
							continue;
						}
						LogicalLine nextLine = LogicalLineParser.ParseLine(st, position, output);

						if (nextLine == null)
							continue;
						if (nextLine is InvalidLine)
						{
							output.PrintWarning(nextLine.ErrMes, position, 2);
						}
						else if (nextLine is FunctionLabelLine)
						{
							FunctionLabelLine label = (FunctionLabelLine)nextLine;
							labelDic.AddLabel(label);
							lastLabelLine = label;
							//if (labelDic.IsOverride(label))
							//    output.PrintWarning("Cxg֐d`Ă܂", position, 0);
						}
						else if (nextLine is GotoLabelLine)
						{
							GotoLabelLine gotoLabel = (GotoLabelLine)nextLine;
							gotoLabel.SetParentFunction(lastLabelLine);
							if (!labelDic.AddLabelDollar((GotoLabelLine)nextLine))
								output.PrintWarning("x$" + line + "͊ɂ̊֐ŎgpĂ܂", position, 2);
						}
						if (lastLabelLine == null)
							output.PrintWarning("֐`Oɍs܂", position, 1);
						else
							lastLine = addLine(nextLine, lastLine);
					}
					addLine(new NullLine(), lastLine);
				}
				finally
				{
					eReader.Close();
				}
				return;
			}


			private LogicalLine addLine(LogicalLine nextLine, LogicalLine lastLine)
			{
				if (nextLine == null)
					return null;
				enabledLineCount++;
				nextLine.PrevLine = lastLine;
				lastLine.NextLine = nextLine;
				return nextLine;
			}

			Dictionary<string, string> renameDic = null;
			internal void LoadEraExRenameFile(string filepath)
			{
				EraStreamReader eReader = new EraStreamReader();
				if ((!File.Exists(filepath)) || (!eReader.Open(filepath)))
				{
					return;
				}

				renameDic = new Dictionary<string, string>();
				string line = null;
				ScriptPosition pos = null;
				try
				{
					while ((line = eReader.ReadLine()) != null)
					{
						if (line.Length == 0)
							continue;
						if (line.StartsWith(";"))
							continue;
						string[] tokens = line.Split(',');
						if (tokens.Length < 2)
							continue;
						pos = new ScriptPosition(eReader.Filename, eReader.LineNo, line);
						//EERB̕\LAϊɂȂB
						string value = tokens[0].Trim();
						string key = string.Format("[[{0}]]", tokens[1].Trim());
						renameDic[key] = value;
						pos = null;
					}
				}
				catch (Exception e)
				{
					if (pos != null)
						throw new CodeEE(e.Message, pos);
					else
						throw new CodeEE(e.Message);

				}
				finally
				{
					eReader.Close();
				}

			}

			private string applyRename(string rowLine)
			{
				if ((rowLine.IndexOf("[[") < 0) || (rowLine.IndexOf("]]") < 0))
					return rowLine;
				if (renameDic == null)
					return rowLine;
				string ret = rowLine;
				foreach (KeyValuePair<string, string> pair in renameDic)
					ret = ret.Replace(pair.Key, pair.Value);
				return ret;
			}

			public bool useCallForm = false;
			/// <summary>
			/// ǍIt@C`FbN
			/// </summary>
			private void checkScript()
			{
				int usedLabelCount = 0;
				int labelDepth = -1;
				List<FunctionLabelLine> labelList = labelDic.GetAllLabels();
				while (true)
				{
					labelDepth++;
					int countInDepth = 0;
					foreach (FunctionLabelLine label in labelList)
					{
						if (label.Depth != labelDepth)
							continue;
						usedLabelCount++;
						countInDepth++;
						checkFunction(label);
					}
					if (countInDepth == 0)
						break;
				}
				labelDepth = -1;
				List<string> ignoredFNCWarningFileList = new List<string>();
				int ignoredFNCWarningCount = 0;

				bool ignoreAll = false;
				switch (Config.Instance.FunctionNotCalledWarning)
				{
					case DisplayWarningFlag.IGNORE:
					case DisplayWarningFlag.LATER:
						ignoreAll = true;
						break;
				}
				if (useCallForm)
				{//callformngꂽSĂ̊֐ĂяoꂽƂ݂ȂB
					foreach (FunctionLabelLine label in labelList)
					{
						if (label.Depth != labelDepth)
							continue;
						checkFunction(label);
					}
				}
				else
				{
					foreach (FunctionLabelLine label in labelList)
					{
						if (label.Depth != labelDepth)
							continue;
						bool ignore = false;
						if (Config.Instance.FunctionNotCalledWarning == DisplayWarningFlag.ONCE)
						{
							string filename = label.Position.Filename.ToUpper();

							if (!string.IsNullOrEmpty(filename))
							{
								if (ignoredFNCWarningFileList.Contains(filename))
								{
									ignore = true;
								}
								else
								{
									ignore = false;
									ignoredFNCWarningFileList.Add(filename);
								}
							}
							break;
						}
						if (ignoreAll || ignore)
							ignoredFNCWarningCount++;
						else
							printWarning("֐@" + label.LabelName + "͒`Ă܂xĂяo܂", label, 1, false);
						if (!Config.Instance.IgnoreUncalledFunction)
							checkFunction(label);
						else
						{
							if (!(label.NextLine is NullLine) && !(label.NextLine is FunctionLabelLine))
							{
								if (!label.NextLine.IsError)
								{
									label.NextLine.IsError = true;
									label.NextLine.ErrMes = "ĂяoȂ͂̊֐Ă΂ꂽ";
								}
							}
						}
					}
				}
				if ((ignoredFNCWarningCount > 0) && (Config.Instance.DisplayWarningLevel <= 1) && (Config.Instance.FunctionNotCalledWarning != DisplayWarningFlag.IGNORE))
					output.PrintLine(string.Format("xLv1:`ꂽ֐xĂяoĂȂɊւx{0}܂", ignoredFNCWarningCount));
				if ((ignoredFNFWarningCount > 0) && (Config.Instance.DisplayWarningLevel <= 2) && (Config.Instance.FunctionNotCalledWarning != DisplayWarningFlag.IGNORE))
					output.PrintLine(string.Format("xLv2:`ĂȂ֐ĂяoɊւx{0}܂", ignoredFNFWarningCount));
				foreach (KeyValuePair<string, int> pair in ignoredWarningCount)
				{
					output.PrintLine(string.Format("xLv{0}ȏ:{1}:{2}̃G[𖳎܂", Config.Instance.DisplayWarningLevel, pair.Key, pair.Value));
				}
				if (Config.Instance.DisplayReport)
					output.PrintLine(string.Format("Ss:{0}, Rgs:{1}, S֐v:{2}, ďo֐v:{3}", lineCount, enabledLineCount, labelDic.Count, usedLabelCount));

			}

			/// <summary>
			/// 
			/// </summary>
			/// <param name="str"></param>
			/// <param name="line"></param>
			/// <param name="level">xx.0:yȃ~X.1:łs.2:ssȂΖQ.3:vI</param>
			private void printWarning(string str, LogicalLine line, int level, bool isError)
			{
				if (isError)
				{
					line.IsError = true;
					line.ErrMes = str;
				}
				if (level < Config.Instance.DisplayWarningLevel)
					return;
				if ((line.Position != null) && (line.Position.Filename != null))
				{
					string filename = line.Position.Filename.ToUpper();
					if (!string.IsNullOrEmpty(filename))
					{
						if (Config.Instance.IgnoreWarningFiles.Contains(filename))
						{
							if (ignoredWarningCount.ContainsKey(filename))
								ignoredWarningCount[filename]++;
							else
								ignoredWarningCount.Add(filename, 1);
							return;
						}
					}
				}
				output.PrintWarning(str, line.Position, level);
			}

			private void printFunctionNotFoundWarning(string str, LogicalLine line, int level, bool isError)
			{
				if (isError)
				{
					line.IsError = true;
					line.ErrMes = str;
				}
				if (level < Config.Instance.DisplayWarningLevel)
					return;
				bool ignore = false;
				if (Config.Instance.FunctionNotFoundWarning == DisplayWarningFlag.IGNORE)
					ignore = true;
				else if (Config.Instance.FunctionNotFoundWarning == DisplayWarningFlag.DISPLAY)
					ignore = false;
				else if (Config.Instance.FunctionNotFoundWarning == DisplayWarningFlag.ONCE)
				{

					string filename = line.Position.Filename.ToUpper();
					if (!string.IsNullOrEmpty(filename))
					{
						if (ignoredFNFWarningFileList.Contains(filename))
						{
							ignore = true;
						}
						else
						{
							ignore = false;
							ignoredFNFWarningFileList.Add(filename);
						}
					}
				}
				if (ignore)
				{
					ignoredFNFWarningCount++;
					return;
				}
				printWarning(str, line, level, isError);
			}

			private void checkFunction(FunctionLabelLine label)
			{
				System.Windows.Forms.Application.DoEvents();
				int depth = label.Depth;
				if (depth < 0)
					depth = -2;
				LogicalLine nextLine = label;
				int ifNest = 0;
				int repeatNest = 0;
				int forNest = 0;
				int whileNest = 0;
                int catchNest = 0;
				string filename = null;
				if (label.Position.Filename != null)
					filename = label.Position.Filename.ToUpper();
				bool ignoreWarning = false;
				if (!string.IsNullOrEmpty(filename))
					ignoreWarning = Config.Instance.IgnoreWarningFiles.Contains(filename);
				while (true)
				{
					nextLine = nextLine.NextLine;
					if ((nextLine is NullLine) || (nextLine is FunctionLabelLine))
						break;
					if (!(nextLine is InstructionLine))
						continue;
					//t[䖽߂̃Wvݒ
					//łIF-ELSEIF-ENDIFAREPEAT-REND̑Ή`FbNȂ
					InstructionLine func = (InstructionLine)nextLine;
					if (Config.Instance.NeedReduceArgumentOnLoad)
					{
						LogicalLineParser.SetArgumentTo(func, this.printWarning);
					}
					LogicalLine jumpto = null;
					switch (func.Function)
					{
						case BuiltInFunctionCode.IF:

							LogicalLine jumptoendif = findEndif(func);
							if (jumptoendif is NullLine)
							{
								printWarning("ΉENDIFȂ܂܃t@CI[ɓB܂", nextLine, 2, true);
								break;
							}
							else if (jumptoendif is FunctionLabelLine)
							{
								printWarning("ΉENDIFȂ܂܊֐I[ɓB܂", nextLine, 2, true);
								break;
							}
							func.JumpToEndif = jumptoendif;
							InstructionLine tempFunc = func;
							jumpto = findIf(tempFunc);
							tempFunc.JumpTo = jumpto;
							while (jumptoendif != jumpto)//ׂĂELSEIFAELSEɃWvݒB
							{
								tempFunc = (InstructionLine)jumpto;
								tempFunc.JumpToEndif = jumptoendif;
								jumpto = findIf(tempFunc);
								tempFunc.JumpTo = jumpto;
							}
							ifNest++;
							break;
						case BuiltInFunctionCode.ELSEIF:
							if ((ifNest <= 0) || (func.JumpToEndif == null))
							{
								printWarning("IF`ENDIF̊OELSEIFg܂", nextLine, 2, true);
								break;
							}
							break;
						case BuiltInFunctionCode.ELSE:
							if ((ifNest <= 0) || (func.JumpToEndif == null))
							{
								printWarning("IF`ENDIF̊OELSEg܂", nextLine, 2, true);
								break;
							}
							jumpto = findIf(func);
							InstructionLine jumptofunc = (InstructionLine)jumpto;
							if (jumptofunc.Function != BuiltInFunctionCode.ENDIF)
							{
								if (jumptofunc.Function == BuiltInFunctionCode.ELSE)
									printWarning("ELSEőELSEgĂ܂", nextLine, 1, false);
								else if (jumptofunc.Function == BuiltInFunctionCode.ELSEIF)
									printWarning("ELSEELSEIFgĂ܂", nextLine, 1, false);
								else
									printWarning("ΉENDIF܂ł", nextLine, 2, true);
							}
							else
								func.JumpTo = jumpto;
							break;
						case BuiltInFunctionCode.ENDIF:
							if (ifNest <= 0)
							{
								printWarning("ΉIF̖ENDIFł", nextLine, 2, true);
								break;
							}
							ifNest--;
							break;
						case BuiltInFunctionCode.REPEAT:
							if (repeatNest >= 1)
							{
								printWarning("REPEATqɂĂ܂", nextLine, 2, true);
								break;
							}
							jumpto = findFunction(func, BuiltInFunctionCode.REND, true);
							if (jumpto is NullLine)
								printWarning("ΉRENDȂ܂܃t@CI[ɓB܂", nextLine, 2, true);
							else if (jumpto is FunctionLabelLine)
								printWarning("ΉRENDȂ܂܊֐I[ɓB܂", nextLine, 2, true);
							else if (jumpto is InstructionLine)
							{
								func.JumpTo = jumpto;
								((InstructionLine)jumpto).JumpTo = func;//RENDREPEAT̈ʒuݒ
								repeatNest++;
							}
							break;
						case BuiltInFunctionCode.REND:
							if ((repeatNest <= 0) || (func.JumpTo == null))
							{
								printWarning("ΉREPEATRENDł", nextLine, 2, true);
								break;
							}
							repeatNest--;
							break;
						case BuiltInFunctionCode.FOR:
							jumpto = findFunction(func, BuiltInFunctionCode.NEXT, true);
							if (jumpto is NullLine)
								printWarning("ΉNEXTȂ܂܃t@CI[ɓB܂", nextLine, 2, true);
							else if (jumpto is FunctionLabelLine)
								printWarning("ΉNEXTȂ܂܊֐I[ɓB܂", nextLine, 2, true);
							else if (jumpto is InstructionLine)
							{
								func.JumpTo = jumpto;
								((InstructionLine)jumpto).JumpTo = func;//NEXTFOR̈ʒuݒ
								forNest++;
							}
							break;
						case BuiltInFunctionCode.NEXT:
							if ((forNest <= 0) || (func.JumpTo == null))
							{
								printWarning("ΉFORNEXTł", nextLine, 2, true);
								break;
							}
							forNest--;
							break;
						case BuiltInFunctionCode.WHILE:
							{
								jumpto = findFunction(func, BuiltInFunctionCode.WEND, true);
								if (jumpto is NullLine)
									printWarning("ΉWENDȂ܂܃t@CI[ɓB܂", nextLine, 2, true);
								else if (jumpto is FunctionLabelLine)
									printWarning("ΉWENDȂ܂܊֐I[ɓB܂", nextLine, 2, true);
								else if (jumpto is InstructionLine)
								{
									func.JumpTo = jumpto;
									((InstructionLine)jumpto).JumpTo = func;//WENDWHILËʒuݒ
									whileNest++;
								}
								break;
							}
						case BuiltInFunctionCode.WEND:
							if ((whileNest <= 0) || (func.JumpTo == null))
							{
								printWarning("ΉWHILEWENDł", nextLine, 2, true);
								break;
							}
							whileNest--;
							break;
						case BuiltInFunctionCode.CONTINUE:
							if ((repeatNest <= 0) && (forNest <= 0) && (whileNest <= 0))
							{
								printWarning("REPEAT`RENDAFOR`NEXTWHILE`WEND̊OCONTINUEg܂", nextLine, 2, true);
								break;
							}
							func.JumpTo = findFunction(func, new BuiltInFunctionCode[] { BuiltInFunctionCode.REPEAT, BuiltInFunctionCode.FOR, BuiltInFunctionCode.WHILE }, false);
							if (func.JumpTo is NullLine)
								printWarning("ΉREPEATAFORAWHILEȂ܂܃t@Cn[ɓB܂", nextLine, 2, true);
							else if (func.JumpTo is FunctionLabelLine)
								printWarning("ΉREPEATAFORAWENDȂ܂܊֐n[ɓB܂", nextLine, 2, true);
							break;
						case BuiltInFunctionCode.BREAK:
							if ((repeatNest <= 0) && (forNest <= 0) && (whileNest <= 0))
							{
								printWarning("REPEAT`RENDAFOR`NEXTWHILE`WEND̊OBREAKg܂", nextLine, 2, true);
								break;
							}
							func.JumpTo = findFunction(func, new BuiltInFunctionCode[] { BuiltInFunctionCode.REND, BuiltInFunctionCode.NEXT, BuiltInFunctionCode.WEND }, true);
							if (func.JumpTo is NullLine)
								printWarning("ΉRENDANEXTAWENDȂ܂܃t@CI[ɓB܂", nextLine, 2, true);
							else if (func.JumpTo is FunctionLabelLine)
								printWarning("ΉRENDANEXTAWENDȂ܂܊֐I[ɓB܂", nextLine, 2, true);
							break;
						case BuiltInFunctionCode.SIF:

							jumpto = func.NextLine;

							if ((jumpto == null) || (jumpto.NextLine == null) ||
							(jumpto is FunctionLabelLine) || (jumpto is NullLine))
								printWarning("SIF̃Xe[gg܂", nextLine, 2, true);
							else if (jumpto is InstructionLine)
							{
								InstructionLine sifFunc = (InstructionLine)jumpto;
								switch (sifFunc.Function)
								{
									case BuiltInFunctionCode.SIF:
									case BuiltInFunctionCode.IF:
									case BuiltInFunctionCode.ELSEIF:
									case BuiltInFunctionCode.ELSE:
									case BuiltInFunctionCode.ENDIF:
									case BuiltInFunctionCode.REPEAT:
									case BuiltInFunctionCode.REND:
									case BuiltInFunctionCode.FOR:
									case BuiltInFunctionCode.NEXT:
									case BuiltInFunctionCode.WHILE:
									case BuiltInFunctionCode.WEND:
										printWarning("SIF̎̍s" + sifFunc.Function.ToString() + "ɂ邱Ƃ͂ł܂", nextLine, 2, true);
										break;
									default:
										func.JumpTo = func.NextLine.NextLine;
										break;
								}
							}
							else
								func.JumpTo = func.NextLine.NextLine;
							break;
						case BuiltInFunctionCode.GOTO://$xփWv
						case BuiltInFunctionCode.TRYGOTO:
                        case BuiltInFunctionCode.TRYCGOTO:
                            {
								string labelName = func.ArgumentStr.Trim();
								if (Config.Instance.IgnoreCase)
									labelName = labelName.ToUpper();
								jumpto = labelDic.GetLabelDollar(labelName, label);
								if ((jumpto == null) && (func.Function == BuiltInFunctionCode.GOTO))
									printWarning("w肳ꂽx\"$" + labelName + "\"݂͌̊֐ɑ݂܂", nextLine, 2, true);
                                else if (jumpto != null)
                                {
                                    func.JumpTo = jumpto;
                                }
                                if (func.Function == BuiltInFunctionCode.TRYCGOTO)
                                {
                                    LogicalLine newNext = findCatch(func, true);
                                    if (newNext is InstructionLine)
                                    {
                                        func.JumpToEndCatch = newNext;
                                        jumpto = findEndCatch(newNext);

                                        if ((jumpto is NullLine) || (jumpto is FunctionLabelLine))
                                        {
                                            printWarning("TRYCGOTOɑΉENDCATCH܂", nextLine, 2, true);
                                            break;
                                        }
                                        ((InstructionLine)newNext).JumpToEndCatch = jumpto;
                                    }
                                    else
                                    {
                                        func.JumpToEndCatch = null;
                                    }
                                    catchNest++;
                                }
                            }
							break;
                        case BuiltInFunctionCode.JUMP:
                        case BuiltInFunctionCode.TRYJUMP:
                        case BuiltInFunctionCode.TRYCJUMP:
                            {
                                if (func.Argument == null)
                                    LogicalLineParser.SetArgumentTo(func, this.printWarning);
                                SpCallArgment callArg = (SpCallArgment)func.Argument;
                                string labelName = callArg.Str;
                                if (Config.Instance.IgnoreCase)
                                    labelName = labelName.ToUpper();
                                FunctionLabelLine targetLabel = labelDic.GetLabel(labelName);
                                if ((targetLabel == null) && (func.Function == BuiltInFunctionCode.JUMP))
                                    printFunctionNotFoundWarning("w肳ꂽ֐\"@" + labelName + "\"݂͑܂", nextLine, 2, true);
                                else if (targetLabel != null)
                                {
                                    func.JumpTo = targetLabel;
                                    if (targetLabel.Depth < 0)
                                        targetLabel.Depth = depth + 1;
                                    checkArgs(targetLabel, callArg.Args, nextLine);
                                }
                                if (func.Function == BuiltInFunctionCode.TRYCJUMP)
                                {
                                    LogicalLine newNext = findCatch(func, true);
                                    if (newNext is InstructionLine)
                                    {
                                        func.JumpToEndCatch = newNext;
                                        jumpto = findEndCatch(newNext);

                                        if ((jumpto is NullLine) || (jumpto is FunctionLabelLine))
                                        {
                                            printWarning("TRYCJUMPɑΉENDCATCH܂", nextLine, 2, true);
                                            break;
                                        }
                                        ((InstructionLine)newNext).JumpToEndCatch = jumpto;
                                    }
                                    else
                                    {
                                        func.JumpToEndCatch = null;
                                    }
                                    catchNest++;
                                }
                             }
                            break;

						case BuiltInFunctionCode.CALL://֐ɈړBړLARETURNŋAB
						case BuiltInFunctionCode.TRYCALL:
                        case BuiltInFunctionCode.TRYCCALL:
                            {
								if (func.Argument == null)
									LogicalLineParser.SetArgumentTo(func, this.printWarning);
								SpCallArgment callArg = (SpCallArgment)func.Argument;
								string labelName = callArg.Str;
								if (Config.Instance.IgnoreCase)
									labelName = labelName.ToUpper();
								FunctionLabelLine targetLabel = labelDic.GetLabel(labelName);
								if ((targetLabel == null) && (func.Function == BuiltInFunctionCode.CALL))
									printFunctionNotFoundWarning("w肳ꂽ֐\"@" + labelName + "\"݂͑܂", nextLine, 2, true);
								else if (targetLabel != null)
								{
									func.JumpTo = targetLabel;
									if (targetLabel.Depth < 0)
										targetLabel.Depth = depth + 1;
									checkArgs(targetLabel, callArg.Args, nextLine);
                                }
                                if (func.Function == BuiltInFunctionCode.TRYCCALL)
                                {
                                    LogicalLine newNext = findCatch(func, true);
                                    if (newNext is InstructionLine)
                                    {
                                        func.JumpToEndCatch = newNext;
                                        jumpto = findEndCatch(newNext);

                                        if ((jumpto is NullLine) || (jumpto is FunctionLabelLine))
                                        {
                                            printWarning("TRYCCALLɑΉENDCATCH܂", nextLine, 2, true);
                                            break;
                                        }
                                        ((InstructionLine)newNext).JumpToEndCatch = jumpto;
                                    }
                                    else
                                    {
                                        func.JumpToEndCatch = null;
                                    }
                                    catchNest++;
                                }
                            }
							break;
						case BuiltInFunctionCode.RESTART:
							func.JumpTo = label;
							break;

						case BuiltInFunctionCode.JUMPFORM:
						case BuiltInFunctionCode.CALLFORM:
						//case BuiltInFunctionCode.GOTOFORM:
						case BuiltInFunctionCode.TRYJUMPFORM:
						case BuiltInFunctionCode.TRYCALLFORM:
							//case BuiltInFunctionCode.TRYGOTOFORM:
                        case BuiltInFunctionCode.TRYCJUMPFORM:
                        case BuiltInFunctionCode.TRYCCALLFORM:
                            useCallForm = true;
                            if (func.Function == BuiltInFunctionCode.TRYCJUMPFORM || func.Function == BuiltInFunctionCode.TRYCCALLFORM)
                            {
                                LogicalLine newNext = findCatch(func, true);
                                if (newNext is InstructionLine)
                                {
                                    func.JumpToEndCatch = newNext;
                                    jumpto = findEndCatch(newNext);

                                    if ((jumpto is NullLine) || (jumpto is FunctionLabelLine))
                                    {
                                        printWarning(func.Function.ToString() + "ɑΉENDCATCH܂", nextLine, 2, true);
                                        break;
                                    }
                                    ((InstructionLine)newNext).JumpToEndCatch = jumpto;
                                    catchNest++;
                                }
                                else
                                {
                                    func.JumpToEndCatch = null;
                                }
                            }
							break;
                        case BuiltInFunctionCode.TRYCGOTOFORM:
                            {
                                LogicalLine newNext = findCatch(func, true);
                                if (newNext is InstructionLine)
                                {
                                    func.JumpToEndCatch = newNext;
                                    jumpto = findEndCatch(newNext);

                                    if ((jumpto is NullLine) || (jumpto is FunctionLabelLine))
                                    {
                                        printWarning("TRYCGOTOFORMCATCHw肵Ă邪AΉENDCATCH܂", nextLine, 2, true);
                                        break;
                                    }
                                    ((InstructionLine)newNext).JumpToEndCatch = jumpto;
                                    catchNest++;
                                }
                                else
                                {
                                    func.JumpToEndCatch = null;
                                }
                            }
                            break;
                        case BuiltInFunctionCode.CATCH:
                            if (catchNest <= 0 || !(findTry(func) is InstructionLine))
                                printWarning("ΉTRYCn֐܂", nextLine, 2, true);
                            else if (!(findEndCatch(func) is InstructionLine))
                                printWarning("ΉENDCATCH܂", nextLine, 2, true);
                            break;
                        case BuiltInFunctionCode.ENDCATCH:
                            if (catchNest <= 0)
                                printWarning("ΉTRYCn֐̂ȂENDCATCHł", nextLine, 2, true);
                            LogicalLine catchLine = findCatch(func, false);
                            if (!(catchLine is InstructionLine))
                                printWarning("ΉCATCĤȂENDCATCHł", nextLine, 2, true);
                            if (!(findTry(catchLine) is InstructionLine))
                                printWarning("TRYCn֐ACATCHAENDCATCHΉĂ܂", nextLine, 2, true);
                            catchNest--;
                            break;
					}
				}
			}

			private bool checkArgs(FunctionLabelLine targetLabel, IOperandTerm[] args, LogicalLine nextLine)
			{
				if (args.Length == 0)
					return true;
				if (targetLabel.Arg.Length < args.Length)
				{
					printWarning("̐֐\"@" + targetLabel.LabelName + "\"ɐݒ肳ꂽ𒴂Ă܂", nextLine, 2, true);
					return false;
				}

				for (int i = 0; i < args.Length; i++)
				{
					if (args[i] == null)
						continue;
					if ((args[i].GetOperandType() == typeof(string)) && (targetLabel.Arg[i].GetOperandType() == typeof(Int64)))
					{
						printWarning((i + 1).ToString() + "Ԗڂ̈̌^֐\"@" + targetLabel.LabelName + "\"ɐݒ肳ꂽ̌^ƈv܂", nextLine, 2, true);
						return false;
					}
				}
				return true;
			}

			//private string splitArgStr(string str)
			//{
			//    string tempStr = null;
			//    StringStream st = new StringStream(str);
			//    TokenReader.SkipWhiteSpace(st);
			//    tempStr = TokenReader.ReadRawString(st);
			//    return tempStr;
			//}

			///elseif,else,endif̂AŏɌ̂ԂB̂݁BqɑΉB
			private LogicalLine findIf(LogicalLine current)
			{
				LogicalLine next = current;
				while (true)
				{
					next = next.NextLine;
					if (next == null)
						throw new ExeEE("IFTɒvIG[");
					if (next is NullLine)
						return next;//t@Cn[͏I[ɓB
					if (next is FunctionLabelLine)
						return next;//֐̃XR[v̊O֏oĂ܂
					if (!(next is InstructionLine))
						continue;
					switch (((InstructionLine)next).Function)
					{
						case BuiltInFunctionCode.ELSE:
						case BuiltInFunctionCode.ELSEIF:
						case BuiltInFunctionCode.ENDIF:
							return next;
						case BuiltInFunctionCode.IF://qIF
							next = findEndif(next);
							if (!(next is InstructionLine))
								return next;
							break;
					}
				}
			}

			///endifԂB̂݁BqɑΉ
			private LogicalLine findEndif(LogicalLine current)
			{
				LogicalLine next = current;
				while (true)
				{
					next = next.NextLine;
					if (next == null)
						throw new ExeEE("");
					if (next is NullLine)
						return next;//t@Cn[͏I[ɓB
					if (next is FunctionLabelLine)
						return next;//֐̃XR[v̊O֏oĂ܂
					if (!(next is InstructionLine))
						continue;
					switch (((InstructionLine)next).Function)
					{
						case BuiltInFunctionCode.ENDIF:
							return next;
						case BuiltInFunctionCode.IF://qIF
							next = findEndif(next);
							if (!(next is InstructionLine))
								return next;
							break;
					}
				}
			}

            private LogicalLine findTry(LogicalLine current)
            {
                LogicalLine prev = current;
                while (true)
                {
                    prev = prev.PrevLine;
                    if (prev == null)
                        throw new ExeEE("");
                    if (prev is FunctionLabelLine)
                        return prev;
                    if (!(prev is InstructionLine))
                        continue;
                    switch (((InstructionLine)prev).Function)
                    {
                        case BuiltInFunctionCode.TRYCCALL:
                        case BuiltInFunctionCode.TRYCCALLFORM:
                        case BuiltInFunctionCode.TRYCGOTO:
                        case BuiltInFunctionCode.TRYCGOTOFORM:
                        case BuiltInFunctionCode.TRYCJUMP:
                        case BuiltInFunctionCode.TRYCJUMPFORM:
                            return prev;
                        case BuiltInFunctionCode.ENDCATCH:
                            prev = findTry(prev);
                            if (!(prev is InstructionLine))
                                return prev;
                            break;
                    }
                }
            }

            private LogicalLine findCatch(LogicalLine current, bool downsearch)
            {
                LogicalLine next = current;
                while (true)
                {
                    if (downsearch)
                        next = next.NextLine;
                    else
                        next = next.PrevLine;
                    if (next == null)
                        throw new ExeEE("");
                    if ((next is FunctionLabelLine) || (next is NullLine))
                        return next;
                    if (next is InstructionLine)
                    {
                        InstructionLine iLine = (InstructionLine)next;
                        //if (iLine.Function == BuiltInFunctionCode.CATCH)
                        //    return next;
                        switch (((InstructionLine)next).Function)
                        {
                            case BuiltInFunctionCode.CATCH:
                                return next;
                            case BuiltInFunctionCode.TRYCCALL:
                            case BuiltInFunctionCode.TRYCCALLFORM:
                            case BuiltInFunctionCode.TRYCGOTO:
                            case BuiltInFunctionCode.TRYCGOTOFORM:
                            case BuiltInFunctionCode.TRYCJUMP:
                            case BuiltInFunctionCode.TRYCJUMPFORM:
                                if (downsearch)
                                {
                                    next = findEndCatch(next);
                                    if (!(next is InstructionLine))
                                        return next;
                                }
                                else
                                    return next; //痈ĂꍇCATCHȂTRY͌킯Ȃ̂ŃG[
                                break;
                            case BuiltInFunctionCode.ENDCATCH:
                                if (downsearch)
                                    return null; //ォ痈ĂꍇCATCHȂENDCATCH͂肦Ȃ
                                else
                                {
                                    next = findTry(next);
                                    if (!(next is InstructionLine))
                                        return next;
                                }
                                break;
                        }
                    }
                }
            }

            private LogicalLine findEndCatch(LogicalLine current)
            {
                LogicalLine next = current;
                while (true)
                {
                    next = next.NextLine;
                    if (next == null)
                        throw new ExeEE("");
                    if (next is NullLine)
                        return next;//t@Cn[͏I[ɓB
                    if (next is FunctionLabelLine)
                        return next;//֐̃XR[v̊O֏oĂ܂
                    if (!(next is InstructionLine))
                        continue;
                    switch (((InstructionLine)next).Function)
                    {
                        case BuiltInFunctionCode.ENDCATCH:
                            return next;
                        case BuiltInFunctionCode.TRYCCALL://qTRY֐Ăяỏ\
                        case BuiltInFunctionCode.TRYCCALLFORM:
                        case BuiltInFunctionCode.TRYCGOTO:
                        case BuiltInFunctionCode.TRYCGOTOFORM:
                        case BuiltInFunctionCode.TRYCJUMP:
                        case BuiltInFunctionCode.TRYCJUMPFORM:
                            next = findEndCatch(next);
                            if (!(next is InstructionLine))
                                return next;
                            break;
                    }
                }
            }

			private LogicalLine findFunction(LogicalLine current, BuiltInFunctionCode findKey, bool downSearch)
			{
				return findFunction(current, new BuiltInFunctionCode[] { findKey }, downSearch);
			}

			/// <summary>
			///@w肵֐TBlXglB
			/// </summary>
			/// <param name="current"></param>
			/// <param name="findKeyArray"></param>
			/// <param name="downSearch"></param>
			/// <returns></returns>
			private LogicalLine findFunction(LogicalLine current, IList<BuiltInFunctionCode> findKeyArray, bool downSearch)
			{
				LogicalLine next = current;
				int forNest = 0;
				int repeatNest = 0;

				while (true)
				{
					if (downSearch)
						next = next.NextLine;
					else
						next = next.PrevLine;
					if (next == null)
						throw new ExeEE("");
					if (next is NullLine)
						return next;//t@Cn[͏I[ɓB
					if (next is FunctionLabelLine)
						return next;//֐̃XR[v̊O֏oĂ܂
					if (!(next is InstructionLine))
						continue;
					InstructionLine iLine = (InstructionLine)next;
					if (downSearch)
					{
						if ((iLine.Function == BuiltInFunctionCode.REPEAT) && (findKeyArray.Contains(BuiltInFunctionCode.REND)))
						{
							repeatNest++;
							continue;
						}
						if ((iLine.Function == BuiltInFunctionCode.FOR) && (findKeyArray.Contains(BuiltInFunctionCode.NEXT)))
						{
							forNest++;
							continue;
						}
					}
					else
					{
						if ((iLine.Function == BuiltInFunctionCode.REND) && (findKeyArray.Contains(BuiltInFunctionCode.REPEAT)))
						{
							repeatNest++;
							continue;
						}
						if ((iLine.Function == BuiltInFunctionCode.NEXT) && (findKeyArray.Contains(BuiltInFunctionCode.FOR)))
						{
							forNest++;
							continue;
						}
					}
					if (findKeyArray.Contains(iLine.Function))
					{
						if ((repeatNest > 0) && ((iLine.Function == BuiltInFunctionCode.REPEAT) || (iLine.Function == BuiltInFunctionCode.REND)))
						{
							repeatNest--;
							continue;
						}
						if ((forNest > 0) && ((iLine.Function == BuiltInFunctionCode.FOR) || (iLine.Function == BuiltInFunctionCode.NEXT)))
						{
							forNest--;
							continue;
						}
						return next;
					}
				}
			}
		}
	}
}
